/**
 * BBMouseGesture - A mouse gesture plugin for Blackbox for Windows
 * Copyright (C) 2004 kana
 * Please read the included file `LICENSE' for license information.
 */

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <BBApi-C.h>
#include <BBCmd_p.h>

#include "cpp_adaptor.h"
#include "debug_console.h"




#ifndef WM_XBUTTONDOWN
#define WM_XBUTTONDOWN 0x020B
#endif
#ifndef WM_XBUTTONUP
#define WM_XBUTTONUP 0x020C
#endif

#ifndef XBUTTON1
#define XBUTTON1 0x0001
#endif
#ifndef XBUTTON2
#define XBUTTON2 0x0002
#endif

#ifndef MOUSEEVENTF_XDOWN
#define MOUSEEVENTF_XDOWN 0x0080
#endif
#ifndef MOUSEEVENTF_XUP
#define MOUSEEVENTF_XUP 0x0100
#endif

#ifndef VK_XBUTTON1
#define VK_XBUTTON1 0x05
#endif
#ifndef VK_XBUTTON2
#define VK_XBUTTON2 0x06
#endif

#ifndef WHEEL_DELTA
#define WHEEL_DELTA 120
#endif
#ifndef GET_WHEEL_DELTA_WPARAM
#define GET_WHEEL_DELTA_WPARAM(wparam) ((short)(HIWORD(wparam)))
#endif

#ifndef DEFINED_MOUSEHOOKSTRUCTEX
#define DEFINED_MOUSEHOOKSTRUCTEX
typedef struct tagMOUSEHOOKSTRUCTEX {
	POINT pt;
	HWND hwnd;
	UINT wHitTestCode;
	DWORD dwExtraInfo;
	DWORD mouseData;
} MOUSEHOOKSTRUCTEX, *LPMOUSEHOOKSTRUCTEX, *PMOUSEHOOKSTRUCTEX;
#endif








typedef enum {
	/* Representing the end of gesture (internal use only) */
	EVENT_END,


	/* Mouse button related events */

	/* Pressing a button */
	EVENT_LDOWN,
	EVENT_RDOWN,
	EVENT_MDOWN,
	EVENT_X1DOWN,
	EVENT_X2DOWN,

	/* Releasing a button */
	EVENT_LUP,
	EVENT_RUP,
	EVENT_MUP,
	EVENT_X1UP,
	EVENT_X2UP,

	/* Moving a wheel */
	EVENT_WHEELUP,
	EVENT_WHEELDOWN,


	/* Cursor movement events */

	/* Moving cursor in that direction */
	EVENT_UP,
	EVENT_DOWN,
	EVENT_LEFT,
	EVENT_RIGHT,

	EVENT_UPLEFT,
	EVENT_UPRIGHT,
	EVENT_DOWNLEFT,
	EVENT_DOWNRIGHT,

	/* Moving cursor to a screen edge */
	EVENT_EDGETOP,
	EVENT_EDGEBOTTOM,
	EVENT_EDGELEFT,
	EVENT_EDGERIGHT,

	EVENT_EDGETOPLEFT,
	EVENT_EDGETOPRIGHT,
	EVENT_EDGEBOTTOMLEFT,
	EVENT_EDGEBOTTOMRIGHT,


	EVENT_NUMBER
} event_id;


typedef unsigned int modifier;
#define MODIFIER_NAN 0
#define MODIFIER_CTRL (1<<0)
#define MODIFIER_META (1<<1)
#define MODIFIER_SHIFT (1<<2)
#define MODIFIER_WIN (1<<3)
#define MODIFIER_UNDER (1<<4)


typedef struct {
	event_id e;  /* the kind of this event */
	modifier m;  /* the modifiers when this event was inputted */

	modifier loose;  /* for checking `ignored modifiers'. */
} event;


/**
 * Return wether the given events are same.
 *
 * Note: The arguments and the right one are distinct.
 *       The left one is treated as an event which is inputted by the user.
 *       The right one is treated as an event which is defined
 *       from the configuration file.
 */
#define EVENT_EQUAL(i,c) \
                (((i).e == (c).e) && (((i).m & ~((c).loose)) == (c).m))


typedef struct {
	event_id e;
	char c;   /* somewhat for debug.  some letters are same, but it's ok. */
	const char* s;
} _event_id_char_string_tuple;

static const _event_id_char_string_tuple EVENT_CHAR_STRING_TABLE[] = {
	{EVENT_END, '$', "END"},

	{EVENT_LDOWN, 'L', "LDown"}, {EVENT_LDOWN, 'L', "L"},
	{EVENT_RDOWN, 'R', "RDown"}, {EVENT_RDOWN, 'R', "R"},
	{EVENT_MDOWN, 'M', "MDown"}, {EVENT_MDOWN, 'M', "M"},
	{EVENT_X1DOWN, 'X', "X1Down"}, {EVENT_X1DOWN, 'X', "X1"},
	{EVENT_X2DOWN, 'Y', "X2Down"}, {EVENT_X2DOWN, 'Y', "X2"},

	{EVENT_LUP, 'l', "LUp"},
	{EVENT_RUP, 'r', "RUp"},
	{EVENT_MUP, 'm', "MUp"},
	{EVENT_X1UP, 'x', "X1Up"},
	{EVENT_X2UP, 'y', "X2Up"},

	{EVENT_WHEELUP, 'w', "WheelUp"},
	{EVENT_WHEELDOWN, 'W', "WheelDown"},

	{EVENT_UP, '8', "Up"},
	{EVENT_DOWN, '2', "Down"},
	{EVENT_LEFT, '4', "Left"},
	{EVENT_RIGHT, '6', "Right"},

	{EVENT_UPLEFT, '7', "UpLeft"},
	{EVENT_UPRIGHT, '9', "UpRight"},
	{EVENT_DOWNLEFT, '1', "DownLeft"},
	{EVENT_DOWNRIGHT, '3', "DownRight"},

	{EVENT_EDGETOP, '8', "EdgeTop"},
	{EVENT_EDGEBOTTOM, '2', "EdgeBottom"},
	{EVENT_EDGELEFT, '4', "EdgeLeft"},
	{EVENT_EDGERIGHT, '6', "EdgeRight"},

	{EVENT_EDGETOPLEFT, '7', "EdgeTopLeft"},
	{EVENT_EDGETOPRIGHT, '9', "EdgeTopRight"},
	{EVENT_EDGEBOTTOMLEFT, '1', "EdgeBottomLeft"},
	{EVENT_EDGEBOTTOMRIGHT, '3', "EdgeBottomRight"},

	{EVENT_END, '\0', NULL},
};

static event_id
event_id_from_string(const char* s)
{
	const _event_id_char_string_tuple* t;

	for (t = EVENT_CHAR_STRING_TABLE; t->s != NULL; t++) {
		if (!strcmp(t->s, s))
			return t->e;
	}

	return EVENT_END;
}

static const char*
event_id_as_string(event_id e)
{
	const _event_id_char_string_tuple* t;

	for (t = EVENT_CHAR_STRING_TABLE; t->s != NULL; t++) {
		if (t->e == e)
			return t->s;
	}

	return NULL;
}

static int
event_id_as_char(event_id e)    /* just for debugging */
{
	const _event_id_char_string_tuple* t;

	for (t = EVENT_CHAR_STRING_TABLE; t->s != NULL; t++) {
		if (t->e == e)
			return t->c;
	}

	return -1;
}

static int
event_id_is_playbackable(event_id e)
{
	switch (e) {
	default:
		return 0;
	case EVENT_LDOWN: case EVENT_LUP:
	case EVENT_RDOWN: case EVENT_RUP:
	case EVENT_MDOWN: case EVENT_MUP:
	case EVENT_X1DOWN: case EVENT_X1UP:
	case EVENT_X2DOWN: case EVENT_X2UP:
	case EVENT_WHEELUP: case EVENT_WHEELDOWN:
		return !0;
	}
}

static void
event_id_as_mouve_event_args(event_id e, DWORD* flags, DWORD* data)
{
	switch (e) {
	default:          *flags = 0; *data = 0; return;
	case EVENT_LDOWN: *flags = MOUSEEVENTF_LEFTDOWN; *data = 0; return;
	case EVENT_LUP:   *flags = MOUSEEVENTF_LEFTUP; *data = 0; return;
	case EVENT_MDOWN: *flags = MOUSEEVENTF_MIDDLEDOWN; *data = 0; return;
	case EVENT_MUP:   *flags = MOUSEEVENTF_MIDDLEUP; *data = 0; return;
	case EVENT_RDOWN: *flags = MOUSEEVENTF_RIGHTDOWN; *data = 0; return;
	case EVENT_RUP:   *flags = MOUSEEVENTF_RIGHTUP; *data = 0; return;
	case EVENT_X1DOWN:*flags = MOUSEEVENTF_XDOWN; *data = XBUTTON1; return;
	case EVENT_X1UP:  *flags = MOUSEEVENTF_XUP; *data = XBUTTON1; return;
	case EVENT_X2DOWN:*flags = MOUSEEVENTF_XDOWN; *data = XBUTTON2; return;
	case EVENT_X2UP:  *flags = MOUSEEVENTF_XUP; *data = XBUTTON2; return;
	case EVENT_WHEELUP:
		*flags = MOUSEEVENTF_WHEEL; *data =  WHEEL_DELTA; return;
	case EVENT_WHEELDOWN:
		*flags = MOUSEEVENTF_WHEEL; *data = -WHEEL_DELTA; return;
	}
}


typedef struct {
	modifier m;
	char c;
	const char* s;
} _modifier_char_string_tuple;

static const _modifier_char_string_tuple MODIFIER_CHAR_STRING_TABLE[] = {
	{MODIFIER_CTRL, 'C', "C"},
	{MODIFIER_CTRL, 'C', "Ctrl"},
	{MODIFIER_CTRL, 'C', "Control"},

	{MODIFIER_META, 'M', "M"},
	{MODIFIER_META, 'M', "Meta"},
	{MODIFIER_META, 'A', "A"},
	{MODIFIER_META, 'A', "Alt"},

	{MODIFIER_SHIFT, 'S', "S"},
	{MODIFIER_SHIFT, 'S', "Shift"},

	{MODIFIER_WIN, 'W', "W"},
	{MODIFIER_WIN, 'W', "Win"},

	{MODIFIER_UNDER, '_', "_"},

	{MODIFIER_NAN, '\0', NULL}
};

static modifier
modifier_from_string(const char* s)
{
	const _modifier_char_string_tuple* t;

	for (t = MODIFIER_CHAR_STRING_TABLE; t->s != NULL; t++)
		if (!strcmp(t->s, s))
			return t->m;

	return MODIFIER_NAN;
}

static modifier
modifier_from_string2(const char* s, int n)
{
	const _modifier_char_string_tuple* t;
	int tn;

	for (t = MODIFIER_CHAR_STRING_TABLE; t->s != NULL; t++) {
		tn = strlen(t->s);
		if (tn != n) continue;

		if (!strncmp(t->s, s, n))
			return t->m;
	}

	return MODIFIER_NAN;
}




/* Maximum length of a gesture */
#define GESTURE_MAX 13

typedef struct {
		/* An additional space for EVENT_END to indicate
		 * the end of a gesture. */
	event gesture[GESTURE_MAX+1];

		/* The command for a gesture */
	BBCmdCommand* command;
} bind;




/**
 * The type to represent a state of BBMouseGesture.
 *
 * For example, the gesture that consists of `R L'
 * is recognized as the following:
 *
 *                         :
 *                         |      STATE_IDLE
 *      Right button down  v ^
 *                           |    STATE_GESTURING
 *      Left  button down    v*^  STATE_ACTION
 *                             |
 *      Right button up        |  STATE_EPILOGUE
 *                             |
 *      Left  button up        v^
 *                              | STATE_IDLE
 *                              :
 *
 * The overlapped arrows mean
 * the state shifting the left one to the right one.
 * The three states
 * -- STATE_GESTURING, STATE_ACTION and STATE_EPILOGUE -- are overlapped,
 * because STATE_ACTION shifts to the next state immediately.
 *
 * STATE_EPILOGUE is a state
 * which avoids passing the end event of a gesture to the application.
 */
typedef enum {
	STATE_IDLE,         /* Waiting a start event */
	STATE_GESTURING,    /* Waiting a end event */
	STATE_ACTION,       /* Execute a command */
	STATE_EPILOGUE      /* ... see the above comment */
} state;




/**
 * Global variables.
 *
 *
 * Term definition
 *
 * Remilia
 *     The process which does something (e.g. loading the configuration,
 *     set/unset the hook and etc) to initialize and finalize BBMouseGesture.
 *     In short a word, the process of the core.
 *     There is only one remilia in the system.
 *     The remilia is a scarlet.
 *
 * Scarlet
 *     The process in which the hook procedure is executed.
 *     There are one or more scarlets in the system.
 *     A scarlet is not necessarily the remilia.
 *
 *
 * Many global variables are shared over many processes.
 * There are two types of global variables:
 *
 * - (64/rw-r--) The variable is written by the remilia
 *               and only read by scarlets.
 * - (66/rw-rw-) The variable is read and written by scarlets.
 *
 * For example, the former is the handle to the hook procedure,
 * the user defined bind or etc, and the latter is
 * the state of BBMouseGesture, the coodinate of the cursor or etc.
 *
 *
 * FIXME: **NO MUTEX**.
 *        I think that the mouse procedure is rarely called by
 *        two or more processes at the same time.
 *        But it's too dangerous.
 */

#define SHARED(X) X __attribute__((section("shared"), shared))

static HINSTANCE SHARED(G_hinstance_self) = NULL;    /* (64) */
static HHOOK SHARED(G_hhook) = NULL;    /* (64) */
static HWND SHARED(G_hwnd_flandre) = NULL;    /* (64) */

static HWND SHARED(G_hwnd_magnolia) = NULL;    /* (60) */

static DWORD SHARED(G_pid_blackbox) = 0;    /* (64) */
static DWORD G_pid_self = 0;    /* (06) -- not shared */

	/* The screen width and height measured in pixel.  (64) */
static int SHARED(G_screen_width) = 0;    /* meaningless value */
static int SHARED(G_screen_height) = 0;    /* meaningless value */

	/* The state of BBMouseGesture.
	 * This will be changed by on_state_xxx only.  (66) */
static state SHARED(G_state) = STATE_IDLE;

	/* The stack contains some inputted events (i.e. a gesture).  (66)
	 * This will not contain EVENT_END to indicate the end of gesture,
	 * so there is no additional space like `bind'.  */
static event SHARED(G_gesture_stack[GESTURE_MAX /*+1-1*/]) = {{EVENT_END}};
static int SHARED(G_gesture_stack_pointer) = -13;    /* meaningless value */

	/* The flags and mask values to represent the mouse state.  (66) */
static unsigned SHARED(G_mouse_state) = 0x0000;
static unsigned SHARED(G_prev_mouse_state) = 0x0000;
#define MS_EDGETOP (1<<0)
#define MS_EDGEBOTTOM (1<<1)
#define MS_EDGELEFT (1<<2)
#define MS_EDGERIGHT (1<<3)
#define MS_LBUTTONDOWN (1<<4)
#define MS_RBUTTONDOWN (1<<5)
#define MS_MBUTTONDOWN (1<<6)
#define MS_X1BUTTONDOWN (1<<7)
#define MS_X2BUTTONDOWN (1<<8)
#define MS_MASK_EDGES (MS_EDGETOP | MS_EDGEBOTTOM | MS_EDGELEFT | MS_EDGERIGHT)
#define MS_MASK_BUTTONDOWNS ( MS_LBUTTONDOWN| MS_RBUTTONDOWN | MS_MBUTTONDOWN \
                             | MS_X1BUTTONDOWN | MS_X2BUTTONDOWN )

	/*The coordinate of the cursor to decide where the cursor moves. (66)*/
static POINT SHARED(G_cursor_point) = {0, 0};    /* meaningless value */

	/* The information of a playbacking gesture.  (66)
	 * This is for avoiding to re-recognize the playbacking gesture. */
static event SHARED(G_playbacking_gesture[GESTURE_MAX /*+1-1*/])={{EVENT_END}};
static int SHARED(G_playbacking_gesture_pointer) = -13; /* meaningless value */
static int SHARED(G_playbacking_gesture_num) = -13;    /* meaningless value */


	/* The start time of an gesture to decide timeout.  (66) */
static DWORD SHARED(G_gesture_start_time) = 0;




	/* The user defined bind table (60) and the size of it.  (64)
	 * This table can not be shared simply,
	 * because G_bind_table is allocate dynamically and
	 * `bind' contains a pointer.
	 * So G_bind_table is owned by the remilia only.
	 * And checking a inputted gesture and executing a command
	 * are delegated to the flandre. */
static bind* G_bind_table = NULL;
static int SHARED(G_bind_table_size) = 0;

	/* Options (64) */

#define OPTDEF_DIAGONAL_CURSOR_MOVEMENTS (0)
static int SHARED(G_option_diagonal_cursor_movements)
	= OPTDEF_DIAGONAL_CURSOR_MOVEMENTS;

#define OPTDEF_PIXEL_THRESHOLD 10
static int SHARED(G_option_pixel_threshold) = OPTDEF_PIXEL_THRESHOLD;

static int SHARED(G_option_start_event[EVENT_NUMBER]) = {
	0,    /* EVENT_END */
	0,    /* EVENT_LDOWN */
	1,    /* EVENT_RDOWN */
	0,    /* EVENT_MDOWN */
	0,    /* EVENT_X1DOWN */
	0,    /* EVENT_X2DOWN */
	0,    /* EVENT_LUP */
	0,    /* EVENT_RUP */
	0,    /* EVENT_MUP */
	0,    /* EVENT_X1UP */
	0,    /* EVENT_X2UP */
	0,    /* EVENT_WHEELUP */
	0,    /* EVENT_WHEELDOWN */
	0,    /* EVENT_UP */
	0,    /* EVENT_DOWN */
	0,    /* EVENT_LEFT */
	0,    /* EVENT_RIGHT */
	0,    /* EVENT_UPLEFT */
	0,    /* EVENT_UPRIGHT */
	0,    /* EVENT_DOWNLEFT */
	0,    /* EVENT_DOWNRIGHT */
	1,    /* EVENT_EDGETOP */
	1,    /* EVENT_EDGEBOTTOM */
	1,    /* EVENT_EDGELEFT */
	1,    /* EVENT_EDGERIGHT */
	0,    /* EVENT_EDGETOPLEFT */
	0,    /* EVENT_EDGETOPRIGHT */
	0,    /* EVENT_EDGEBOTTOMLEFT */
	0,    /* EVENT_EDGEBOTTOMRIGHT */
};
static int SHARED(G_option_end_event[EVENT_NUMBER]) = {
	0,    /* EVENT_END */
	1,    /* EVENT_LDOWN */
	1,    /* EVENT_RDOWN */
	1,    /* EVENT_MDOWN */
	1,    /* EVENT_X1DOWN */
	1,    /* EVENT_X2DOWN */
	1,    /* EVENT_LUP */
	1,    /* EVENT_RUP */
	1,    /* EVENT_MUP */
	1,    /* EVENT_X1UP */
	1,    /* EVENT_X2UP */
	0,    /* EVENT_WHEELUP */
	0,    /* EVENT_WHEELDOWN */
	0,    /* EVENT_UP */
	0,    /* EVENT_DOWN */
	0,    /* EVENT_LEFT */
	0,    /* EVENT_RIGHT */
	0,    /* EVENT_UPLEFT */
	0,    /* EVENT_UPRIGHT */
	0,    /* EVENT_DOWNLEFT */
	0,    /* EVENT_DOWNRIGHT */
	0,    /* EVENT_EDGETOP */
	0,    /* EVENT_EDGEBOTTOM */
	0,    /* EVENT_EDGELEFT */
	0,    /* EVENT_EDGERIGHT */
	0,    /* EVENT_EDGETOPLEFT */
	0,    /* EVENT_EDGETOPRIGHT */
	0,    /* EVENT_EDGEBOTTOMLEFT */
	0,    /* EVENT_EDGEBOTTOMRIGHT */
};
static int SHARED(G_option_single_event[EVENT_NUMBER]) = {
	0,    /* EVENT_END */
	0,    /* EVENT_LDOWN */
	0,    /* EVENT_RDOWN */
	0,    /* EVENT_MDOWN */
	0,    /* EVENT_X1DOWN */
	0,    /* EVENT_X2DOWN */
	0,    /* EVENT_LUP */
	0,    /* EVENT_RUP */
	0,    /* EVENT_MUP */
	0,    /* EVENT_X1UP */
	0,    /* EVENT_X2UP */
	0,    /* EVENT_WHEELUP */
	0,    /* EVENT_WHEELDOWN */
	0,    /* EVENT_UP */
	0,    /* EVENT_DOWN */
	0,    /* EVENT_LEFT */
	0,    /* EVENT_RIGHT */
	0,    /* EVENT_UPLEFT */
	0,    /* EVENT_UPRIGHT */
	0,    /* EVENT_DOWNLEFT */
	0,    /* EVENT_DOWNRIGHT */
	1,    /* EVENT_EDGETOP */
	1,    /* EVENT_EDGEBOTTOM */
	1,    /* EVENT_EDGELEFT */
	1,    /* EVENT_EDGERIGHT */
	0,    /* EVENT_EDGETOPLEFT */
	0,    /* EVENT_EDGETOPRIGHT */
	0,    /* EVENT_EDGEBOTTOMLEFT */
	0,    /* EVENT_EDGEBOTTOMRIGHT */
};

#define OPTDEF_TIMEOUT 1000
static int SHARED(G_option_timeout) = OPTDEF_TIMEOUT;

#define OPTDEF_EDGE_WIDTH 10
static int SHARED(G_option_edge_width) = OPTDEF_EDGE_WIDTH;

#define OPTDEF_CORNER_WIDTH 50
static int SHARED(G_option_corner_width) = OPTDEF_CORNER_WIDTH;

#define OPTDEF_NAVIGATION_WINDOW (0)
static int SHARED(G_option_navigation_window) = OPTDEF_NAVIGATION_WINDOW;

#define OPTDEF_ONLY_BLACKBOX (!0)
static int SHARED(G_option_only_blackbox) = OPTDEF_ONLY_BLACKBOX;




static const char* const BBMOUSEGESTURE_NAME = "BBMouseGesture";
static const char* const BBMOUSEGESTURE_VERSION = "0.1.6";
static const char* const BBMOUSEGESTURE_AUTHOR = "kana";
static const char* const BBMOUSEGESTURE_RELEASE = "2004-03-12/10-09";
static const char* const BBMOUSEGESTURE_LINK = "http://nicht.s8.xrea.com/";
static const char* const BBMOUSEGESTURE_EMAIL = "nicht AT s8 DOT xrea DOT com";
static const char* const BBMOUSEGESTURE_DEFAULT = __FILE__;


static const char* const FLANDRE_CLASSNAME = "BBMouseGesture/Flandre";
static const char* const FLANDRE_WINDOWNAME = "BBMouseGesture/Flandre";

#define FLANDRE_WM_EXECUTE (WM_APP+0x0013)


static const char* const MAGNOLIA_CLASSNAME = "BBMouseGesture/Magnolia";
                         /* "BBMouseGesture/Magnolia"; -- see [2004-07-20] */
static const char* const MAGNOLIA_WINDOWNAME = NULL;

#define MAGNOLIA_INIT (WM_APP+'i')
#define MAGNOLIA_INPUT (WM_APP+'I')
#define MAGNOLIA_EXECUTE (WM_APP+'E')
#define MAGNOLIA_NOTFOUND (WM_APP+'N')
#define MAGNOLIA_TIMEOUT (WM_APP+'T')
#define MAGNOLIA_UPDATE (WM_APP+'U')




#define arraysizeof(a) (sizeof(a) / sizeof((a)[0]))

#define SETFLAG(f,v) ((f) |= (v))
#define UNSETFLAG(f,v) ((f) &= ~(v))


#define is_mouse_button_reversed() GetSystemMetrics(SM_SWAPBUTTON)


/* Return the string corresponding to a given error code.
 * The caller must free the returning value by LocalFree. */
static LPVOID
get_error_message(DWORD error_code)
{
	LPVOID s;

	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER
	                   | FORMAT_MESSAGE_FROM_SYSTEM
	                   | FORMAT_MESSAGE_IGNORE_INSERTS,
	               NULL,
	               error_code,
	               MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
	               (LPTSTR)&s,
	               0,
	               NULL );

	return s;
}


static void
make_always_on_top(HWND hwnd, int flag)
{
	SetWindowPos( hwnd,
	              (flag ? HWND_TOPMOST : HWND_NOTOPMOST),
	              0x4C, 0x69, 0x74, 0x79,
	              SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER
	                | SWP_NOSIZE | SWP_SHOWWINDOW );
}




static int
is_existent(const char* const path_to_a_file)
{
	FILE* f;

	f = fopen(path_to_a_file, "r");
	if (f == NULL) return 0;

	fclose(f);

	return !0;
}


/**
 * Return the string corresponding to a given message (WM_XXX).  For debug.
 *
 * BUGS: This doesn't convert all kinds of message.
 * BUGS: This returns static array.
 */
static const char* const
message_id_to_string(WPARAM wp)
{
	switch (wp) {
	case WM_LBUTTONUP: return "WM_LBUTTONUP";
	case WM_LBUTTONDOWN: return "WM_LBUTTONDOWN";
	case WM_RBUTTONUP: return "WM_RBUTTONUP";
	case WM_RBUTTONDOWN: return "WM_RBUTTONDOWN";
	case WM_MBUTTONUP: return "WM_MBUTTONUP";
	case WM_MBUTTONDOWN: return "WM_MBUTTONDOWN";
	case WM_MOUSEMOVE: return "WM_MOUSEMOVE";
	case WM_CAPTURECHANGED: return "WM_CAPTURECHANGED";
	case WM_LBUTTONDBLCLK: return "WM_LBUTTONDBLCLK";
	case WM_RBUTTONDBLCLK: return "WM_RBUTTONDBLCLK";
	case WM_MBUTTONDBLCLK: return "WM_MBUTTONDBLCLK";
	case WM_MOUSEACTIVATE: return "WM_MOUSEACTIVATE";
	case WM_MOUSEHOVER: return "WM_MOUSEHOVER";
	case WM_MOUSELEAVE: return "WM_MOUSELEAVE";
	case WM_MOUSEWHEEL: return "WM_MOUSEWHEEL";
	case WM_NCHITTEST: return "WM_NCHITTEST";
	case WM_XBUTTONUP: return "WM_XBUTTONUP";
	case WM_XBUTTONDOWN: return "WM_XBUTTONDOWN";
	default: {
		static char buf[40];
		snprintf(buf, arraysizeof(buf), "0x%08X", wp);
		return buf;
		}
	}
}




/* Output log */

/* level */
#define LO_EMERG        0x00
#define LO_ALERT        0x01
#define LO_CRIT         0x02
#define LO_ERR          0x03
#define LO_WARNING      0x04
#define LO_NOTICE       0x05
#define LO_INFO         0x06
#define LO_DEBUG        0x07

#define LO_MASK_LEVEL   0x07

/* where to output */
#define LO_CONS         (0x01<<3)    /* show as dialog */
#define LO_DECONS       (0x02<<3)    /* send to debug_console */
#define LO_FILE         (0x04<<3)    /* write to file */

#ifndef BUILD_AS_RELEASE_VERSION   /* debug */
#define LO_DEFAULT      (LO_DECONS)
#define LO_THRESHOLD    LO_DEBUG
#else    /* release */
#define LO_DEFAULT      0
#define LO_THRESHOLD    LO_WARNING
#endif


void
lo(int flag, const char* const format, ...)
{
	va_list ap;
	char buf[80*2+1];

	flag |= LO_DEFAULT;

	if (LO_THRESHOLD < (flag & LO_MASK_LEVEL)) return;

	va_start(ap, format);
	vsnprintf(buf, arraysizeof(buf), format, ap);
	va_end(ap);

	if (flag & LO_CONS) {
		UINT icon;

		icon = ((flag & LO_MASK_LEVEL) < LO_ERR)
		       ? MB_ICONINFORMATION : MB_ICONERROR;
		MessageBox( NULL, buf, BBMOUSEGESTURE_NAME,
			    icon | MB_OK | MB_APPLMODAL | MB_SETFOREGROUND );
	}
	if (flag & LO_DECONS) {
		HWND console;
		COPYDATASTRUCT cds;

		console = FindWindow( DEBUG_CONSOLE_CLASSNAME,
		                      DEBUG_CONSOLE_WINDOWNAME );
		if (console == NULL) return;

		cds.dwData = DEBUG_CONSOLE_WM_NOTICE;
		cds.cbData = strlen(buf) + 1;
		cds.lpData = (PVOID)buf;

		SendMessage(console, WM_COPYDATA, (WPARAM)NULL,(LPARAM)(&cds));
	}
	if (flag & LO_FILE) {
		; /* FIXME: Not implemented yet */
	}
}




/**
 * Find the path to the configuration file and store it to `path'.
 * This doesn't store more than `path_size' character
 * (including last NUL).
 *
 * Find the file in the following order:
 *     1. PATH_TO_BBMG/BBMouseGesture.rc
 *     2. PATH_TO_BBMG/BBMouseGesturerc
 *     3. PATH_TO_CORE/BBMouseGesture.rc
 *     4. PATH_TO_CORE/BBMouseGesturerc
 */
static int
find_path_to_configuration_file(char* path, int path_size)
{
	HINSTANCE hinstances[] = {G_hinstance_self, NULL};
	const char* const filenames[] = { "BBMouseGesture.rc",
	                                  "BBMouseGesturerc" };
	char dir[1024];    /* FIXME */
	char* p;
	int i;
	int j;

	lo(LO_DEBUG, "load_setting");
	for (i = 0; i < arraysizeof(hinstances); i++) {
		if (!GetModuleFileName(hinstances[i], dir, arraysizeof(dir)))
			continue;
		dir[arraysizeof(dir)-1] = '\0';

		p = dir + strlen(dir) - 1;
		while (dir <= p) {
			if (*p == '\\')
				break;
			p--;
		}
		*(p + 1) = '\0';

		for (j = 0; j < arraysizeof(filenames); j++) {
			snprintf( path, path_size,
			          "%s%s", dir, filenames[j] );
			path[path_size-1] = '\0';
			lo(LO_DEBUG, "load_setting: try to load: `%s'", path);
			if (is_existent(path))
				return !0;
		}
	}

	return 0;
}




/**
 * Functions to operate G_gesture_stack.
 *
 * - The bottom of G_gesture_stack is 0.
 * - G_gesture_stack_pointer points the most top element of the stack.
 *   For example, -1 means empty, 0 means there is 1 element...
 */
static void
gesture_stack_init(void)
{
	G_gesture_stack_pointer = -1;
}

/* Just ignore if the stack is full. */
static int
gesture_stack_push(const event* g)
{
	if ((int)arraysizeof(G_gesture_stack) - 1 <= G_gesture_stack_pointer)
		return 0;

	G_gesture_stack_pointer++;
	G_gesture_stack[G_gesture_stack_pointer] = *g;

	if (G_gesture_stack_pointer == 0)
		SendMessage(G_hwnd_magnolia, MAGNOLIA_INIT, 0, 0);
	SendMessage(G_hwnd_magnolia, MAGNOLIA_INPUT, g->m, g->e);

	return !0;
}

/* Not used
static int
gesture_stack_pop(void)
{
	if (G_gesture_stack_pointer < 0)
		return 0;

	G_gesture_stack_pointer--;

	return !0;
}
*/

static const event*
gesture_stack_top(void)
{
	if (G_gesture_stack_pointer < 0)
		return NULL;

	return G_gesture_stack + G_gesture_stack_pointer;
}

static int
gesture_stack_size(void)
{
	return G_gesture_stack_pointer + 1;
}




/**
 * Play back a given `events' which must be end with EVENT_END.
 * Any unplaybackable (!event_id_is_playbackable(e)) events are just ignored.
 *
 * This will copy `events' to 'G_playbacking_gesture'
 * for avoiding to re-recognize the playbacking gesture (== `events').
 * See also mouse_procedure.
 *
 * BUGS: Don't pass many events more than GESTURE_MAX,
 *       It will be played back correctly,
 *       but the preventing re-recognizing algorithm
 *       will not work correctly (this is not a big problem though).
 *
 * FIXME: The implementation is using mouse_event with the assumption that
 *        any events will not be occured while executing this.
 *        For the future,
 *        this should be implemented with SendInput if it is available.
 *        Because generating events with SendInput can not be intercepted.
 *
 * BUGS: The caller which is a LowLevelMouseProc (or etc) must *NOT* call this.
 *       Calling by the procedures causes an endless loop,
 *       because they will be called when the system posts a message into
 *       a thread message queue.  [2004-03-22T14:23]
 */

static DWORD
reverse_mouseeventf(DWORD f)
{
	if (f == MOUSEEVENTF_LEFTDOWN) return MOUSEEVENTF_RIGHTDOWN;
	else if (f == MOUSEEVENTF_LEFTUP) return MOUSEEVENTF_RIGHTUP;
	else if (f == MOUSEEVENTF_RIGHTDOWN) return MOUSEEVENTF_LEFTDOWN;
	else if (f == MOUSEEVENTF_RIGHTUP) return MOUSEEVENTF_LEFTUP;
	else return f;
}

/**
 * Regularize a given `f' which is a MOUSEEVENTF_XXX.
 *
 * mouse_event doesn't consider
 * whether the logical mouse buttons are reverse to the physical mouse button.
 */
static DWORD
regularize_mouseeventf(DWORD f)
{
	return (!is_mouse_button_reversed()) ? f : reverse_mouseeventf(f);
}


int
modifier_as_virtualkey(modifier m)
{
	switch (m) {
	default: return -1;
	case MODIFIER_CTRL: return VK_CONTROL;
	case MODIFIER_META: return VK_MENU;
	case MODIFIER_SHIFT: return VK_SHIFT;
	case MODIFIER_WIN: return VK_LWIN;
	}
}


static void
_playback_an_event(const event* e)    /* assuming `e' is playbackable */
{
	const modifier MODIFIERS[] = {
		MODIFIER_CTRL,
		MODIFIER_META,
		MODIFIER_SHIFT,
		MODIFIER_WIN,
		MODIFIER_NAN
	};
	const modifier* m;
	DWORD flag;
	DWORD data;

	for (m = MODIFIERS; *m != MODIFIER_NAN; m++) {
	  if (e->m & *m)
	    keybd_event(modifier_as_virtualkey(*m), 0, 0, 0);
	   else
	    keybd_event(modifier_as_virtualkey(*m), 0, KEYEVENTF_KEYUP, 0);
	}

	event_id_as_mouve_event_args(e->e, &flag, &data);
	mouse_event(regularize_mouseeventf(flag), 0, 0, data, 0);

	for (m = MODIFIERS; *m != MODIFIER_NAN; m++) {
	  if (e->m & *m)
	    keybd_event(modifier_as_virtualkey(*m), 0, KEYEVENTF_KEYUP, 0);
	}
}

static void
playback_event(const event* events)
{
	const event* e;

	G_playbacking_gesture_pointer = 0;
	G_playbacking_gesture_num = 0;

	for (e = events; e->e != EVENT_END; e++) {
		if (!event_id_is_playbackable(e->e)) continue;

		_playback_an_event(e);

		if (G_playbacking_gesture_num < GESTURE_MAX) {
			G_playbacking_gesture[G_playbacking_gesture_num] = *e;
			G_playbacking_gesture_num++;
		}
	}
}





/* The functions corresponding to each state of BBMouseGesture. */

static int on_state_gesturing(const event*, HWND);
static int on_state_action(const event*, HWND);
static int on_state_epilogue(const event*, HWND);

static void
change_state(state s, HWND window)
{
	switch (s) {
	default:
		break;
	case STATE_IDLE:
		G_state = STATE_IDLE;
		gesture_stack_init();
		ReleaseCapture();
		break;
	case STATE_GESTURING:
		G_state = STATE_GESTURING;
		G_gesture_start_time = GetTickCount();
		SetCapture(window);
		break;
	case STATE_ACTION:
		G_state = STATE_ACTION;
		break;
	case STATE_EPILOGUE:
		G_state = STATE_EPILOGUE;
		break;
	}
}


static int
on_state_idle(const event* e, HWND window)
{
	if (!(G_option_start_event[e->e])) return 0;
	if ( G_option_only_blackbox && (G_pid_self != G_pid_blackbox)
	     && (e->e < EVENT_UP) ) /* Is a button event */
		return 0;

	gesture_stack_push(e);

	if (!(G_option_single_event[e->e])) {
		change_state(STATE_GESTURING, window);
		return !0;
	} else {
		change_state(STATE_ACTION, window);
		return on_state_action(e, window);
	}
}


static int
on_state_gesturing(const event* e, HWND window)
{
	if (G_option_single_event[e->e]) return !0;

	if (!(G_option_end_event[e->e])) {
		switch (e->e) {
		default:
			break;

		/* To avoid inputting a same direction event */
		case EVENT_UP:       case EVENT_DOWN:
		case EVENT_LEFT:     case EVENT_RIGHT:
		case EVENT_UPLEFT:   case EVENT_UPRIGHT:
		case EVENT_DOWNLEFT: case EVENT_DOWNRIGHT:
			if (e->e == gesture_stack_top()->e)
				return !0;
			break;
		}

		gesture_stack_push(e);
		return !0;
	} else {
		gesture_stack_push(e);
		change_state(STATE_ACTION, window);
		return on_state_action(e, window);
	}
}


static int
on_state_action(const event* e, HWND window)
{
	char s[arraysizeof(G_gesture_stack)*2 + 1];
	int i;

	for (i = 0; i <= G_gesture_stack_pointer; i++) {
		s[i*2 + 0] = ' ';
		s[i*2 + 1] = event_id_as_char(G_gesture_stack[i].e);
	}
	s[i*2] = '\0';

	lo( LO_DEBUG|LO_DECONS, "   state %c  sp%2d [%s ]",
	    "IGAE"[G_state], G_gesture_stack_pointer, s );

	SendMessage(G_hwnd_flandre, FLANDRE_WM_EXECUTE, 0, 0);

	change_state(STATE_EPILOGUE, window);
	return on_state_epilogue(e, window);
}


static int
on_state_epilogue(const event* e, HWND window)
{
	if (!(G_mouse_state & (MS_MASK_BUTTONDOWNS | MS_MASK_EDGES))) {
		change_state(STATE_IDLE, window);
	}

	return !0;
}




/**
 * Load / unload configuration
 * ```````````````````````````
 *
 * BUGS: Loading and unloading modify the variables about the configurations,
 *       so these MUST NOT be called while executing the hook procedure.
 *       Therefore, BBMouseGesture can not reload the configuration
 *       by reconfiguring the core instead of restarting.
 */

#define DEFAULT_CONF \
	( "gesture *C-*M-*S-*W-LDown  *C-*M-*S-*W-LUp  = playback()\n" \
	  "gesture *C-*M-*S-*W-RDown  *C-*M-*S-*W-RUp  = playback()\n" \
	  "gesture *C-*M-*S-*W-MDown  *C-*M-*S-*W-MUp  = playback()\n" \
	  "gesture *C-*M-*S-*W-X1Down *C-*M-*S-*W-X1Up = playback()\n" \
	  "gesture *C-*M-*S-*W-X2Down *C-*M-*S-*W-X2Up = playback()\n" )


/* tokenizer */

typedef enum {
	TI_END,

	TI_GESTURE,
	TI_OPTION,

	TI_SCRIPT,
	TI_NAME,
	TI_NUMBER,
	TI_BOOL,

	TI_EQUAL,
	TI_MINUS,
	TI_ASTER,
	TI_TILDE,
	TI_NL
} token_id;

typedef struct {
	token_id id;
	char* attr;
	int line;
	int column;
} token;

typedef struct {
	char* _mem;
	int num;
	token* tokens;
	int current;
} toquence;




/* Check keywords only.  They are parsed as TI_NAME. */

typedef struct {
	token_id id;
	const char* s;
} _token_id_string_tuple;

static _token_id_string_tuple _TOKEN_ID_STRING_TABLE[] = {
	{TI_GESTURE, "gesture"},
	{TI_OPTION, "option"},
	{TI_BOOL, "false"},
	{TI_BOOL, "true"},
	{TI_END, NULL},
};

static token_id
token_id_from_string(const char* s)
{
	const _token_id_string_tuple* p;

	for (p = _TOKEN_ID_STRING_TABLE; p->id != TI_END; p++) {
		if (!strncmp(s, p->s, strlen(p->s)))
			return p->id;
	}

	return TI_NAME;
}


static int isfirstletterofname(int c)
	{return ('a'<=c && c<='z') || ('A'<=c && c<='Z') || (c=='_');}
static int issecondletterofname(int c)
	{return isfirstletterofname(c) || ('0' <= c && c <= '9');}
static int isbinarydigit(int c) {return '0' == c || c == '1';}
static int isdecimaldigit(int c) {return '0' <= c && c <= '9';}
static int isoctaldigit(int c) {return '0' <= c && c <= '7';}
static int ishexaldigit(int c)
{
	return    ('0' <= c && c <= '9')
	       || ('a' <= c && c <= 'f') || ('A' <= c && c <= 'F');
}


/**
 * Return a sequence of tokens from a given `string'.
 * If this failed, then return NULL
 * and store the information in `*err_line', `*err_column', `*err_message'.
 *
 * err_line, err_column and err_message must be a valid pointer to
 * an existent object (i.e. must not be NULL).
 */
static toquence*
toquence_from_string( const char* string,
                      int* err_line,
                      int* err_column,
                      const char** err_message )
{
	toquence* self;

	char* p;
	char* attr;
	token_id id;
	token* tmp;
	int lineno;
	char* first_column;
	char* token_column;
	token_id current_stmt = TI_END;

	self = (toquence*)malloc(sizeof(*self));
		if (self == NULL) {
			*err_line = 0;
			*err_column = 0;
			*err_message = "Out of memory (toquence)";
			goto e_oom_toquence;
		}
	self->current = 0;
	self->num = 0;
	self->tokens = NULL;
	self->_mem = strdup(string);
		if (self->_mem == NULL) {
			*err_line = 0;
			*err_column = 0;
			*err_message = "Out of memory (toquence->_mem)";
			goto e_oom_strdup;
		}

	lineno = 1;  first_column = token_column = self->_mem;
	p = self->_mem;
	while (*p != '\0') {
		if (current_stmt == TI_GESTURE && id == TI_EQUAL) {
			attr = p;
			token_column = p;

			do {
				++p;
			} while (*p != '\n' && *p != '\0');

			id = TI_SCRIPT;
			goto skip_normal_process;
		}

		switch (*p) {
		/* WS */
		case ' ':
		case '\t':
			*p = '\0';  p++;
			continue;
		/* COMMENT */
		case '#':
			*p = '\0';  p++;
			while (*p != '\n' && *p != '\0')
				p++;
			continue;
		/* Others */
		case '\n':id = TI_NL;        attr = NULL;
		          token_column = p;  *p = '\0'; p++;  break;
		case '=': id = TI_EQUAL;     attr = NULL;
		          token_column = p;  *p = '\0'; p++;  break;
		case '-': id = TI_MINUS;     attr = NULL;
		          token_column = p;  *p = '\0'; p++;  break;
		case '*': id = TI_ASTER;     attr = NULL;
		          token_column = p;  *p = '\0'; p++;  break;
		case '~': id = TI_TILDE;     attr = NULL;
		          token_column = p;  *p = '\0'; p++;  break;
		/* For the future: bindmap.
		case '\'':
			token_column = p;
			*p = '\0';  p++;
			attr = p;

			while (!0) {
				if (*p == '\'' || *p == '\0') break;
				if (*p == '\n') {lineno++; first_column = p+1;}
				p++;
			}
			if (*p != '\'') {
				err_message = "Unterminated string";
				goto e_unterminated_string;
			}
			*p = '\0';
			p++;
			id = TI_STRING;
			break;
		case '"': {
			char* q;

			token_column = p;
			*p = '\0';  p++;
			attr = q = p;

			while (!0) {
				if (*p == '"' || *p == '\0') break;
				if (*p == '\n') {lineno++; first_column = p+1;}

				if (*p == '\\') {
					p++;
					if (*p != '\\' && *p != '"') {
					  err_message
					    = "Invalid char in string";
					  goto e_invalid_char_in_string;
					}
				}

				*q = *p;  p++;  q++;
			}
			if (*p != '"') {
				err_message = "Unterminated string";
				goto e_unterminated_string;
			}
			p++;
			*q = '\0';
			id = TI_STRING;
			} break;
		*/
		default:
			token_column = p;
			if (isfirstletterofname(*p)) {
				char* q;

				attr = q = p;

				do {p++;} while (issecondletterofname(*p));

				id = token_id_from_string(attr);

				if (id == TI_GESTURE || id == TI_OPTION)
					current_stmt = id;
			} else if (isdecimaldigit(*p)) {
				int (*isdigit)(int);
				attr = p;
				if ('0' == *p) {
					p++;
					if (*p == 'x' || *p == 'X')
						{p++; isdigit = ishexaldigit;}
					else if (*p == 'o' || *p == 'O')
						{p++; isdigit = isoctaldigit;}
					else if (*p == 'b' || *p == 'B')
						{p++; isdigit = isbinarydigit;}
					else
						{isdigit = isoctaldigit;}
				} else
					isdigit = isdecimaldigit;
				while ((*isdigit)(*p))
					p++;
				if ( isdecimaldigit(*p)
				     || isfirstletterofname(*p) ) {
					*err_message ="Invalid char in number";
					goto e_invalid_char_in_number;
				}
				id = TI_NUMBER;
			} else {
				*err_message = "Invalid char";
				goto e_invalid_char;
			}
			break;
		}
skip_normal_process:
		tmp = (token*)realloc( self->tokens,
		                       sizeof(token) * (self->num+1) );
		if (tmp == NULL) goto e_oom_realloc;
		self->tokens = tmp;
		self->tokens[self->num].id = id;
		self->tokens[self->num].attr = attr;
		self->tokens[self->num].column = (token_column - first_column)
		                                      / sizeof(char) + 1;
		self->tokens[self->num].line = lineno;
		self->num++;
		if (id == TI_NL) {lineno++;  first_column = p+1; }
	}

	return self;

e_oom_realloc:
e_invalid_char_in_string:
e_unterminated_string:
e_invalid_char:
e_invalid_char_in_number:
	*err_line = lineno;
	*err_column = (p - first_column) / sizeof(char) + 1;
	free(self->tokens);
	free(self->_mem);
e_oom_strdup:
	free(self);
e_oom_toquence:
	return NULL;
}

static void
toquence_del(toquence* self)
{
	free(self->_mem);
	free(self->tokens);
	free(self);
}

/* Return true for no more tokens */
static int
toquence_end(const toquence* self)
{
	return !(self->current < self->num);
}

/* Skip to next token */
static void
toquence_next(toquence* self)
{
	self->current++;
}

/* Return the current token */
static const token*
toquence_current(const toquence* self)
{
	return toquence_end(self) ? NULL : (self->tokens + self->current);
}

/* Return the last token  */
static const token*
toquence_last(const toquence* self)
{
	return (0 <= self->num - 1) ? (self->tokens + self->num - 1) : NULL;
}

/* Return true for the id of the current token matching a given id */
static int
toquence_match(toquence* self, token_id id)
{
	return (!toquence_end(self)) && (self->tokens[self->current].id == id);
}

/* Skip current line (to recover an error) */
static void
toquence_panic(toquence* self)
{
	while (!toquence_end(self) && !toquence_match(self, TI_NL))
		toquence_next(self);
	toquence_next(self);
}

/* Short for */
static int
toquence_matchnext(toquence* self, token_id id)
{
	if (toquence_match(self, id)) {
		toquence_next(self);
		return !0;
	} else {
		return 0;
	}
}


static void
notice_error(const char* file, int line, int column, const char* message)
{
	int n;
	const char* p;

	p = DEFAULT_CONF;
	n = 0;
	while (*p != '\0') {
		if (*p == '\n')
			n++;
		p++;
	}

	lo(LO_ERR|LO_CONS, "%s:%d:%d: %s", file, line - n, column, message);
}

static void
parsing_error(const char* message, const char* file, const toquence* toq)
{
	const token* tok;

	tok = toquence_end(toq) ? toquence_last(toq) : toquence_current(toq);

	notice_error(file, tok->line, tok->column, message);
}




/**
 * gesture_statement: 'gesture' event+ '=' script nl
 * event: mod* event_name
 * mod: ('~' | '*')? mod_name '-'
 */
static int
rule_gesture( toquence* toq, const char* conf_file,
              const BBCmdCompiler* compiler )
{
	event gesture[arraysizeof(G_gesture_stack) + 1];
	int i;
	const char* script;
	BBCmdCommand* command;
	bind* t;
	int error_line;
	int error_column;
	const char* error_message;

	lo(LO_DEBUG, "--------------");
	if (!toquence_matchnext(toq, TI_GESTURE)) goto ERR;

	i = 0;
	while (i < arraysizeof(gesture) - 1) {
		int mod_type;
		const char* a;
		modifier m;

		gesture[i].m = 0;
		gesture[i].loose = 0;

		while (!0) {
			mod_type = ' ';
			if (toquence_matchnext(toq, TI_TILDE))
				mod_type = '~';
			else if (toquence_matchnext(toq, TI_ASTER))
				mod_type = '*';

			if (!toquence_match(toq, TI_NAME)) goto ERR;
			a = toquence_current(toq)->attr;
			toquence_next(toq);

			if (!toquence_matchnext(toq, TI_MINUS)) break;

			m = modifier_from_string(a);
			if (m == MODIFIER_NAN || m == MODIFIER_UNDER) goto ERR;

			if (mod_type == ' ') {
				gesture[i].m     |=  m;
				gesture[i].loose &= ~m;
			} else if (mod_type == '~') {
				gesture[i].m     &= ~m;
				gesture[i].loose &= ~m;
			} else if (mod_type == '*') {
				gesture[i].m     &= ~m;
				gesture[i].loose |=  m;
			} else {
			  lo( LO_ERR|LO_CONS,
			      "Something wrong at rule_gesture (mod_type: %X)",
			      mod_type );
			}
		}

		gesture[i].e = event_id_from_string(a);
		if (gesture[i].e == EVENT_END) goto ERR;

		i++;

		if (toquence_match(toq, TI_EQUAL))
			break;
	}

	gesture[i].e = EVENT_END;
	if (!toquence_matchnext(toq, TI_EQUAL)) goto ERR;
	if (!toquence_match(toq, TI_SCRIPT)) goto ERR;
	script = toquence_current(toq)->attr;
	toquence_next(toq);
	if (!toquence_matchnext(toq, TI_NL)) goto ERR;

	command = BBCmdCompiler_CompileFromString
	              ( compiler, script,
	                &error_line, &error_column, &error_message );
	if (command == NULL) {
		parsing_error(error_message, conf_file, toq);
		/* Don't toquence_panic here -- already skipped */
		return 0;
	}

	t = realloc(G_bind_table, sizeof(*G_bind_table)*(G_bind_table_size+1));
	if (t == NULL) {
		parsing_error("Out of memory", conf_file, toq);
		BBCmdCommand_Del(command);
		/* Don't toquence_panic here -- already skipped */
		return 0;
	}
	for (i = 0; gesture[i].e != EVENT_END; i++) {
		lo(LO_DEBUG, "%02X/%02X/%s", gesture[i].m, gesture[i].loose, event_id_as_string(gesture[i].e));
		t[G_bind_table_size].gesture[i] = gesture[i];
	}
	t[G_bind_table_size].gesture[i] = gesture[i];
	t[G_bind_table_size].command = command;
	G_bind_table = t;
	G_bind_table_size++;

	return !0;

ERR:
	parsing_error("Syntax error", conf_file, toq);
	toquence_panic(toq);
	return 0;
}

static int
rule_option(toquence* toq, const char* conf_file)
{
	const char* name;
	int events[EVENT_NUMBER];
	int i;
	int type;

	if (!toquence_matchnext(toq, TI_OPTION)) goto ERR;

	if (!toquence_match(toq, TI_NAME)) goto ERR;
	name = toquence_current(toq)->attr;
	toquence_next(toq);

	if (!toquence_matchnext(toq, TI_EQUAL)) goto ERR;
	if (toquence_end(toq)) goto ERR;
	switch (toquence_current(toq)->id) {
	case TI_NUMBER:
		type = 'n';
		i = atoi(toquence_current(toq)->attr);
		toquence_next(toq);
		break;
	case TI_BOOL:
		type = 'b';
		i = (toquence_current(toq)->attr[0] == 't');
		toquence_next(toq);
		break;
	default:
		type = 'e';
		memset(events, 0, sizeof(events));
		i = 0;
		while ( toquence_match(toq, TI_NAME)
		        && i < arraysizeof(events) - 1 )
		{
			event_id e;

			e = event_id_from_string(toquence_current(toq)->attr);
			if (e == EVENT_END) goto ERR;

			events[e] = !0;

			i++;
			toquence_next(toq);
		}
		break;
	}
	if (!toquence_matchnext(toq, TI_NL)) goto ERR;

	if (!strcmp(name, "start_event") && type == 'e') {
		memcpy(G_option_start_event, events, sizeof(events));
	} else if (!strcmp(name, "end_event") && type == 'e') {
		memcpy(G_option_end_event, events, sizeof(events));
	} else if (!strcmp(name, "single_event") && type == 'e') {
		memcpy(G_option_single_event, events, sizeof(events));
	} else if (!strcmp(name, "diagonal_cursor_movements") && type == 'b') {
		G_option_diagonal_cursor_movements = i;
	} else if (!strcmp(name, "pixel_threshold") && type == 'n' && 0 < i) {
		G_option_pixel_threshold = i;
	} else if (!strcmp(name, "timeout") && type == 'n' && 0 < i) {
		G_option_timeout = i;
	} else if (!strcmp(name, "edge_width") && type == 'n' && 0 <= i) {
		G_option_edge_width = i;
	} else if (!strcmp(name, "corner_width") && type == 'n' && 0 <= i) {
		G_option_corner_width = i;
	} else if (!strcmp(name, "navigation_window") && type=='b' && 0 <= i) {
		G_option_navigation_window = i;
	} else if (!strcmp(name, "only_blackbox") && type=='b' && 0 <= i) {
		G_option_only_blackbox = i;
	} else {
		goto ERR;
	}

	return !0;

ERR:
	parsing_error("Syntax error", conf_file, toq);
	toquence_panic(toq);
	return 0;
}




/**
 * `playback' - a BBCmdCommand to play back some events
 *
 * For example:
 *
 * - playback('L', 'LUp') will do left click
 * - playback('C-L', 'C-LUp') will do left click with ctrl modifier
 * - playback('_-L', '_-LUp') will do left click with any modifiers
 *   which are pressed when you input a gesture.
 * - playback('_-~C-L', '_-~C-LUp') will do like the above,
 *   but this ignore ctrl modifier.
 *   Note that '~C-_-L' doesn't work as the above,
 *   because specifying modifiers are evaluated left to right
 *   and `_' will override the settings of the left modifiers.
 *
 * BUGS: `playback' cannot process `*Mod'.
 *       Using `_' instead of `*Mod'.
 *
 *       This command uses pl_events[i].m to remember
 *       what modifier this should play back,
 *       and uses pl_events[i].loose to remember
 *       what modifier this should not play back.
 */

typedef struct _BBCmdPlaybackCommand {
	BBCmdCommand_HEAD

	event* pl_events;    /* terminated by EVENT_END */
	event* pl_swap;    /* working space to process `_' modifier */
} BBCmdPlaybackCommand;

/* forward */
static const BBCmdCommandDescriptor _BBCmdPlaybackCommandDescriptor;
#define BBCmdPlaybackCommandDescriptor (&_BBCmdPlaybackCommandDescriptor)

static const char*
_BBCmdPlaybackCommand_initialize( BBCmdCommand* _self,
                                  int ac, const char* const* av,
                                  const BBCmdCompiler* compiler )
{
	BBCmdPlaybackCommand* self;
	int i;
	event e;
	event* t;
	int n;

	assert(_self->cmd_descriptor == BBCmdPlaybackCommandDescriptor);
	self = (BBCmdPlaybackCommand*)_self;

	n = 0;
	self->pl_events = NULL;
	self->pl_swap = NULL;

	for (i = 0 + 1; i < ac; i++) {
		const char* s;
		size_t len;
		modifier m;
		int mod_type;

		e.m = 0;
		e.loose = 0;

		s = av[i];
		while (!0) {
			if (s[0] == '~') {mod_type = s[0];  s++;}
			else mod_type = ' ';

			len = strcspn(s, "-");
			if (s[len] == '\0') {
			  if (mod_type == ' ') break;
			  else return "`~' may be prefixed"
			              " modifier only (not event)";
			}

			m = modifier_from_string2(s, len);
			if (m == MODIFIER_NAN)
				return "Unknown modifier";
			if (m == MODIFIER_UNDER && mod_type == '~')
				return "Unable to understand `~_'";

			if (m == MODIFIER_UNDER) {
				e.m      =  0;
				e.m     |=  MODIFIER_UNDER;
				e.loose  =  0;
			} else if (mod_type == ' ') {
				e.m     |=  m;
				e.loose &= ~m;
			} else if (mod_type == '~') {
				e.m     &= ~m;
				e.loose |=  m;
			} else {
				lo( LO_ERR|LO_CONS,
				    "Something wrong at"
				    " _BBCmdPlaybackCommand_initialize"
				    " (mod_type: %X)",
				    mod_type );
			}

			s += len + 1;
		}

		e.e = event_id_from_string(s);
		if (e.e == EVENT_END) return "Unknown event";

		t = (event*)realloc(self->pl_events, sizeof(*t) * (n + 1));
		if (t == NULL) return "Out of memory";

		self->pl_events = t;
		self->pl_events[n] = e;
		n++;
	}

	t = (event*)realloc(self->pl_events, sizeof(*t) * (n + 1));
	if (t == NULL) return "Out of memory";

	self->pl_events = t;
	self->pl_events[n].e = EVENT_END;
	n++;

	self->pl_swap = (event*)malloc(sizeof(self->pl_swap[0]) * n);
	if (self->pl_swap == NULL) return "Out of memory";
	memcpy(self->pl_swap, self->pl_events, sizeof(self->pl_swap[0]) * n);

	return NULL;
}

static void
_BBCmdPlaybackCommand_del(BBCmdCommand* _self)
{
	BBCmdPlaybackCommand* self;

	assert(_self->cmd_descriptor == BBCmdPlaybackCommandDescriptor);
	self = (BBCmdPlaybackCommand*)_self;

	free(self->pl_events);
	free(self->pl_swap);
}

static int
_BBCmdPlaybackCommand_execute(const BBCmdCommand* _self)
{
	BBCmdPlaybackCommand* self;
	int i;

	assert(_self->cmd_descriptor == BBCmdPlaybackCommandDescriptor);
	self = (BBCmdPlaybackCommand*)_self;

	if (self->pl_events[0].e != EVENT_END) {
		for (i = 0; self->pl_events[i].e != EVENT_END; i++) {
		  if (self->pl_events[i].m & MODIFIER_UNDER) {
		    if (i < gesture_stack_size()) {
		      self->pl_swap[i].m  = G_gesture_stack[i].m;
		      self->pl_swap[i].m |=   self->pl_events[i].m;
		      self->pl_swap[i].m &= ~(self->pl_events[i].loose);
		    } else {
		      self->pl_swap[i].m = self->pl_events[i].m;
		    }
		  }
		}

		playback_event(self->pl_swap);
	} else {
		event events[arraysizeof(G_gesture_stack) + 1];

		memcpy(events, G_gesture_stack, sizeof(G_gesture_stack));
		events[gesture_stack_size()].e = EVENT_END;

		playback_event(events);
	}

	return BBCmd_RC_TRUE;
}


static const BBCmdCommandDescriptor _BBCmdPlaybackCommandDescriptor = {
	"playback",
	NULL,
	"([event [, event [...]]])\n"
		"Play back given `event'.\n"
		"If `event' is omitted,\n"
		"then play back the gesture which is inputted just before.\n",
	sizeof(BBCmdPlaybackCommand),
	_BBCmdPlaybackCommand_initialize,
	_BBCmdPlaybackCommand_del,
	_BBCmdPlaybackCommand_execute
};




/* The main code of loading */

static int
load_setting(void)
{
	BBCmdCompiler* compiler;
	char path[1024];
	FILE* file;
	long file_size;
	toquence* toq;
	char* conf;
	int err_line;
	int err_column;
	const char* err_message;
	int return_code = 0;


	G_bind_table_size = 0;
	G_bind_table = NULL;

	compiler = BBCmdCompiler_New();
	if (compiler == NULL) {
		lo(LO_ERR | LO_CONS, "Out of memory (BBCmdCompiler_New)");
		goto ERR1;
	}
	if (!BBCmdCompiler_AddNewCommand( compiler,
	                                  BBCmdPlaybackCommandDescriptor )) {
		lo( LO_ERR | LO_CONS,
		    "Out of memory (BBCmdCompiler_AddNewCommand)" );
		goto ERR2;
	}

	if (!find_path_to_configuration_file(path, arraysizeof(path))) {
		lo(LO_WARNING | LO_CONS, "Configuration file not found");
		goto ERR2;
	}

	lo(LO_DEBUG, "load_setting: loading `%s'", path);
	file = fopen(path, "rt");
	if (file == NULL) {
		lo(LO_ERR | LO_CONS, "Unable to open `%s' for reading", path);
		goto ERR2;
	}

	if (fseek(file, 0L, SEEK_END) != 0) {
		lo(LO_ERR|LO_CONS, "Unable to read `%s' (fseek failed)", path);
		goto ERR3;
	}
	file_size = ftell(file);
	if (file_size == -1) {
		lo(LO_ERR|LO_CONS, "Unable to read `%s' (ftell failed)", path);
		goto ERR3;
	}

	conf = (char*)malloc( sizeof(char)
	                      * (strlen(DEFAULT_CONF) + file_size + 1) );
	if (conf == NULL) {
		lo(LO_ERR|LO_CONS, "Out of memory (load_setting/malloc)");
		goto ERR3;
	}

	rewind(file);
	strcpy(conf, DEFAULT_CONF);
	if ( fread(conf + strlen(DEFAULT_CONF), sizeof(char), file_size, file)
	     != file_size )
	{
		lo(LO_ERR|LO_CONS, "Unable to read `%s' (fread failed)",path);
		goto ERR4;
	}
	conf[strlen(DEFAULT_CONF) + file_size] = '\0';

	toq = toquence_from_string(conf, &err_line, &err_column, &err_message);
	if (toq == NULL) {
		notice_error(path, err_line, err_column, err_message);
		goto ERR4;
	}

	while (!toquence_end(toq)) {
		switch (toquence_current(toq)->id) {
		case TI_GESTURE:
			if (!rule_gesture(toq, path, compiler)) goto ERR5;
			break;
		case TI_OPTION:
			if (!rule_option(toq, path)) goto ERR5;
			break;
		case TI_NL:
			toquence_next(toq);
			break;
		default:
			parsing_error("Syntax error", path, toq);
			toquence_panic(toq);
			goto ERR5;
			break;
		}
	}

	{
	  event_id e;

	  for (e = EVENT_END; e < EVENT_NUMBER; e++) {
	    G_option_start_event[e]
	      = (G_option_start_event[e] || G_option_single_event[e]);
	  }
	}

	lo( LO_DEBUG, "option diagonal_cursor_movements: %s",
	    (G_option_diagonal_cursor_movements ? "enable" : "disable") );
	lo(LO_DEBUG, "option pixel_threshold: %d", G_option_pixel_threshold);
	lo(LO_DEBUG, "option timeout: %d", G_option_timeout);
	lo(LO_DEBUG, "option edge_width: %d", G_option_edge_width);
	lo(LO_DEBUG, "option corner_width: %d", G_option_corner_width);

	return_code = !0;

ERR5:
	toquence_del(toq);
ERR4:
	free(conf);
ERR3:
	fclose(file);
ERR2:
	BBCmdCompiler_Del(compiler);
ERR1:
	lo(LO_DEBUG, "load_setting: end");

	return return_code;
}


static void
unload_setting(void)
{
	int i;

	lo(LO_DEBUG, "unload_setting");

	for (i = 0; i < G_bind_table_size; i++)
		BBCmdCommand_Del(G_bind_table[i].command);
	free(G_bind_table);
	G_bind_table = NULL;
	G_bind_table_size = 0;
}




static LRESULT CALLBACK
flandre_window_procedure(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	int i;
	int j;

	if (msg != FLANDRE_WM_EXECUTE) return DefWindowProc(hwnd, msg, wp, lp);

	/* Search the user-defined binds from last to first
	 * to override a bind by another which is defined later. */
	for (i = G_bind_table_size - 1; 0 <= i; i--) {
		for ( j = 0;
		      j < gesture_stack_size()
			  && G_bind_table[i].gesture[j].e != EVENT_END;
		      j++ )
		{
			if (!EVENT_EQUAL( G_gesture_stack[j],
			                  G_bind_table[i].gesture[j] ))
			{
				break;
			}
		}
		if ( j < gesture_stack_size()
		     || G_bind_table[i].gesture[j].e != EVENT_END )
		{
			continue;
		}

		BBCmdCommand_Execute(G_bind_table[i].command);
		break;
	}

	SendMessage( G_hwnd_magnolia,
	             (0 <= i ? MAGNOLIA_EXECUTE : MAGNOLIA_NOTFOUND), 0, 0 );

	return TRUE;
}




static event_id
decide_cursor_movement_direction(POINT p)
{
	/**
	 *        Y
	 *        ^
	 *     7 6|2 3 
	 *   5    |    1
	 *   4   O|    0
	 * -------+------>X
	 *   C    |    8
	 *   D    |    9
	 *     F E|A B
	 *        |
	 *
	 * Vertical direction of the values in TABLE
	 * is reverse to the above figure,
	 * because the screen origin (0, 0) is top-left corner
	 * and increase to bottom-right.
	 */
	static const event_id TABLE[0xF + 1] = {
		EVENT_RIGHT,
		EVENT_DOWNRIGHT,
		EVENT_DOWN,
		EVENT_DOWNRIGHT,
		EVENT_LEFT,
		EVENT_DOWNLEFT,
		EVENT_DOWN,
		EVENT_DOWNLEFT,

		EVENT_RIGHT,
		EVENT_UPRIGHT,
		EVENT_UP,
		EVENT_UPRIGHT,
		EVENT_LEFT,
		EVENT_UPLEFT,
		EVENT_UP,
		EVENT_UPLEFT,
	};
	POINT d;
	POINT a;
	int swap = 0;
	event_id e;

	d.x = p.x - G_cursor_point.x;
	d.y = p.y - G_cursor_point.y;
	a.x = (d.x < 0) ? -d.x : d.x;
	a.y = (d.y < 0) ? -d.y : d.y;

	/**
	 * Square threshold (old)
	 *    a.x < G_option_pixel_threshold && a.y < G_option_pixel_threshold
	 */

	/* Circle threshold */
	if ( a.x * a.x + a.y * a.y
	     < G_option_pixel_threshold * G_option_pixel_threshold )
		return EVENT_END;

	if (a.x < a.y) {LONG t = a.x;  a.x = a.y;  a.y = t; swap = !0;}

	G_cursor_point.x = p.x;
	G_cursor_point.y = p.y;

	/* now 0 <= atan(a.y/a.x) <= 1 */
	/* tan(pi/8) = 0.4142135... */

		/* is diagonal movement? */
	if (414 < 1000*a.y / a.x && !G_option_diagonal_cursor_movements)
		return EVENT_END;

	e = TABLE[   (414 < 1000*a.y / a.x ? 0x1 : 0x0)
		   | (swap                 ? 0x2 : 0x0)
		   | (d.x < 0              ? 0x4 : 0x0)
		   | (d.y < 0              ? 0x8 : 0x0) ];

	return e;
}


static event_id
decide_cursor_shiftin_edge(POINT p)
{
	static const event_id table[0x0F+1] = {
		EVENT_END,              /* ____ */
		EVENT_EDGETOP,          /* ___T */
		EVENT_EDGEBOTTOM,       /* __B_ */
		EVENT_END,              /* __BT */
		EVENT_EDGELEFT,         /* _L__ */
		EVENT_EDGETOPLEFT,      /* _L_T */
		EVENT_EDGEBOTTOMLEFT,   /* _LB_ */
		EVENT_EDGELEFT,         /* _LBT */
		EVENT_EDGERIGHT,        /* R___ */
		EVENT_EDGEBOTTOMLEFT,   /* R__T */
		EVENT_EDGEBOTTOMRIGHT,  /* R_B_ */
		EVENT_EDGERIGHT,        /* R_BT */
		EVENT_END,              /* RL__ */
		EVENT_EDGETOP,          /* RL_T */
		EVENT_EDGEBOTTOM,       /* RLB_ */
		EVENT_END               /* RLBT */
	};
	unsigned s;

	s = G_mouse_state ^ G_prev_mouse_state;
	s &= MS_MASK_EDGES;
	s &= G_mouse_state;

	return table[s];
}


/**
 * Return whether `pt' is in an edge or a corner of the screen.
 *
 *    +-------------------
 *    |corner|    edge
 *    |  +----------------
 *    |  |
 *    |--|
 *    |  |
 *    |  |
 *    |edge
 *    |  |
 *    |  |
 */
static int
check_point(POINT pt)
{
  unsigned e;
  unsigned c;

  e = (
      ((pt.y <= G_option_edge_width - 1)               ? MS_EDGETOP    : 0)
    | ((G_screen_height - G_option_edge_width <= pt.y) ? MS_EDGEBOTTOM : 0)
    | ((pt.x <= G_option_edge_width - 1)               ? MS_EDGELEFT   : 0)
    | ((G_screen_width - G_option_edge_width <= pt.x)  ? MS_EDGERIGHT  : 0)
  );

  c = (
      ((pt.y <= G_option_corner_width - 1)               ? MS_EDGETOP    : 0)
    | ((G_screen_height - G_option_corner_width <= pt.y) ? MS_EDGEBOTTOM : 0)
    | ((pt.x <= G_option_corner_width - 1)               ? MS_EDGELEFT   : 0)
    | ((G_screen_width - G_option_corner_width <= pt.x)  ? MS_EDGERIGHT  : 0)
  );

  if (e != 0)
    return e | c;
  else
    return 0;
}


/**
 * If this function translates WM_NCXXXBUTTONYYY messeages to EVENT_XXXYYY,
 * then any menus of BB4W will be unoperatable.
 * Because they receives WM_NCXXXBUTTONYYY, not WM_XXXBUTTONYYY.
 * And there is no easy way to play back WM_NCXXXBUTTONYYY messages.
 * mouse_event and SendInput are not able to them.
 *
 * So BBMouseGesture can not recognize a gesture which
 * starts with a button down event and it occurred in non client area.
 *
 * 0.1.3: Now this function never change G_mouse_state
 *        because it's not changed sometimes.
 */
static event_id
event_id_from_windows_message(UINT message, const MOUSEHOOKSTRUCTEX* mhsex)
{
	G_prev_mouse_state = G_mouse_state;

	switch (message) {
	default: break;

	case WM_LBUTTONDOWN:
		return EVENT_LDOWN;
	case WM_LBUTTONUP:
		return EVENT_LUP;
	case WM_MBUTTONDOWN:
		return EVENT_MDOWN;
	case WM_MBUTTONUP:
		return EVENT_MUP;
	case WM_RBUTTONDOWN:
		return EVENT_RDOWN;
	case WM_RBUTTONUP:
		return EVENT_RUP;
	case WM_XBUTTONDOWN:
		switch (mhsex->mouseData) {
		default: return EVENT_END;
		case XBUTTON1: return EVENT_X1DOWN;
		case XBUTTON2: return EVENT_X2DOWN;
		}
		break;
	case WM_XBUTTONUP:
		switch (mhsex->mouseData) {
		default: return EVENT_END;
		case XBUTTON1: return EVENT_X1UP;
		case XBUTTON2: return EVENT_X2UP;
		}
		break;

	case WM_MOUSEWHEEL:
		if (0 < GET_WHEEL_DELTA_WPARAM(mhsex->mouseData))
			return EVENT_WHEELUP;
		else if (GET_WHEEL_DELTA_WPARAM(mhsex->mouseData) < 0)
			return EVENT_WHEELDOWN;
		else
			return EVENT_END;

	case WM_MOUSEMOVE:    case WM_NCMOUSEMOVE:
	case WM_MOUSEHOVER:/* case WM_NCMOUSEHOVER: where is this defined? */
	case WM_MOUSELEAVE:/* case WM_NCMOUSELEAVE: where is this defined? */ {
		event_id e_m;
		event_id e_e;

		G_mouse_state &= ~MS_MASK_EDGES;
		G_mouse_state |= check_point(mhsex->pt);

		e_m = decide_cursor_movement_direction(mhsex->pt);
		e_e = decide_cursor_shiftin_edge(mhsex->pt);

		switch ((e_m!=EVENT_END ? 1 : 0) | (e_e!=EVENT_END ? 2 : 0)) {
		case 0:
			break;
		case 1:
			return e_m;
		case 2:
			return e_e;
		case 3: /* FIXME: Which value should this return? */
			return e_e;
		default:
			abort();
		}

		} break;
	}

	return EVENT_END;
}


#define USHORT_MSB (1 << (sizeof(USHORT)*CHAR_BIT - 1))

static modifier
modifier_from_getkeystate(void)
{
	modifier m;

	m = 0;

	if (GetKeyState(VK_CONTROL) & USHORT_MSB)
		m |= MODIFIER_CTRL;
	if (GetKeyState(VK_MENU) & USHORT_MSB)
		m |= MODIFIER_META;
	if (GetKeyState(VK_SHIFT) & USHORT_MSB)
		m |= MODIFIER_SHIFT;
	if (GetKeyState(VK_LWIN) & USHORT_MSB)
		m |= MODIFIER_WIN;
	if (GetKeyState(VK_RWIN) & USHORT_MSB)
		m |= MODIFIER_WIN;

	return m;
}




static void
check_mouse_button_state(void)
{
	if (GetAsyncKeyState(VK_LBUTTON) & USHORT_MSB)
		SETFLAG(G_mouse_state,   MS_LBUTTONDOWN);
	else
		UNSETFLAG(G_mouse_state, MS_LBUTTONDOWN);

	if (GetAsyncKeyState(VK_RBUTTON) & USHORT_MSB)
		SETFLAG(G_mouse_state,   MS_RBUTTONDOWN);
	else
		UNSETFLAG(G_mouse_state, MS_RBUTTONDOWN);

	if (GetAsyncKeyState(VK_MBUTTON) & USHORT_MSB)
		SETFLAG(G_mouse_state,   MS_MBUTTONDOWN);
	else
		UNSETFLAG(G_mouse_state, MS_MBUTTONDOWN);

	if (GetAsyncKeyState(VK_XBUTTON1) & USHORT_MSB)
		SETFLAG(G_mouse_state,   MS_X1BUTTONDOWN);
	else
		UNSETFLAG(G_mouse_state, MS_X1BUTTONDOWN);

	if (GetAsyncKeyState(VK_XBUTTON2) & USHORT_MSB)
		SETFLAG(G_mouse_state,   MS_X2BUTTONDOWN);
	else
		UNSETFLAG(G_mouse_state, MS_X2BUTTONDOWN);
}




static HWND
new_flandre(HINSTANCE hinstance)
{
	WNDCLASS wc;
	HWND hwnd;

	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = flandre_window_procedure;
	wc.hInstance = hinstance;
	wc.lpszClassName = FLANDRE_CLASSNAME;
	if (!RegisterClass(&wc)) return NULL;

	hwnd = CreateWindowEx( WS_EX_TOOLWINDOW,
	                       FLANDRE_CLASSNAME,
	                       FLANDRE_WINDOWNAME,
	                       WS_ICONIC,
	                       0, 0, 0, 0,
	                       HWND_MESSAGE,
	                       NULL,
	                       hinstance,
	                       NULL );
	if (hwnd == NULL) {UnregisterClass(FLANDRE_CLASSNAME, hinstance);}

	return hwnd;
}


static void
del_flandre(HWND hwnd, HINSTANCE hinstance)
{
	DestroyWindow(hwnd);
	UnregisterClass(FLANDRE_CLASSNAME, hinstance);
}




typedef void* (*tGetSettingPtr)(int);
static FARPROC SHARED(pGetBBWnd) = NULL;
static FARPROC SHARED(pMakeSticky) = NULL;
static FARPROC SHARED(pRemoveSticky) = NULL;
static FARPROC SHARED(pGetSettingPtr) = NULL;
static FARPROC SHARED(pMakeStyleGradient) = NULL;
static FARPROC SHARED(pCreateStyleFont) = NULL;
static FARPROC SHARED(pSnapWindowToEdge) = NULL;
static void* SHARED(G_style) = NULL;
static const int SUBSCRIBING_MSGS[] = {BB_RECONFIGURE, BB_BROADCAST, 0};

static char SHARED(G_text[256]) = "";

static int
is_pressed(int key)
{
	return GetKeyState(key) & (1 << (sizeof(SHORT) * CHAR_BIT - 1));
}

static LRESULT CALLBACK
magnolia_window_procedure(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{

	switch (msg) {
	default:
		break;

	case MAGNOLIA_INIT:
		G_text[0] = '\0';
		break;
	case MAGNOLIA_INPUT: {
		char* p;

		p = G_text + strlen(G_text);
		if (G_text != p)  /* strlen(G_text) != 0 */
			*(p++) = ' ';
		if ((modifier)wp & MODIFIER_CTRL) {*(p++) = 'C'; *(p++) = '-';}
		if ((modifier)wp & MODIFIER_META) {*(p++) = 'M'; *(p++) = '-';}
		if ((modifier)wp & MODIFIER_SHIFT){*(p++) = 'S'; *(p++) = '-';}
		if ((modifier)wp & MODIFIER_WIN)  {*(p++) = 'W'; *(p++) = '-';}
		*(p++) = event_id_as_char((event_id)lp);
		*p = '\0';
		} goto _update;
	case MAGNOLIA_EXECUTE:
		strcat(G_text, " -> !");
		goto _update;
	case MAGNOLIA_NOTFOUND:
		strcat(G_text, " -> undefined");
		goto _update;
	case MAGNOLIA_TIMEOUT:
		strcat(G_text, " -> timeout");
		goto _update;
	_update:
	case MAGNOLIA_UPDATE: {
		RECT r;

		GetWindowRect(hwnd, &r);
		SetWindowPos( hwnd, NULL, -13, -13,
		              24 + strlen(G_text)
		                     * _StyleItem_FontHeight(G_style) / 2,
		              r.bottom - r.top,
		              SWP_NOMOVE|SWP_NOACTIVATE|SWP_NOOWNERZORDER
		                |SWP_NOZORDER );

		InvalidateRect(hwnd, NULL, TRUE);
		} break;

	case BB_RECONFIGURE:
		G_style = (*(tGetSettingPtr)pGetSettingPtr)(SN_TOOLBAR);
		InvalidateRect(hwnd, NULL, TRUE);
		break;

	case BB_BROADCAST: {
		LPCSTR s = (LPCSTR)lp;

		if (!stricmp(s, "@BBShowPlugins"))
			ShowWindow(hwnd, SW_SHOWNA);
		else if (!stricmp(s, "@BBHidePlugins"))
			ShowWindow(hwnd, SW_HIDE);
		else
			; /* nothing */

		} break;

	case WM_CREATE: {
		#define FIND(proc) GetProcAddress(GetModuleHandle(NULL), proc)
		pGetBBWnd = FIND("GetBBWnd");
		pMakeSticky = FIND("MakeSticky");
		pRemoveSticky = FIND("RemoveSticky");
		pGetSettingPtr = FIND("GetSettingPtr");
		pMakeStyleGradient = FIND("MakeStyleGradient");
		pCreateStyleFont = FIND("CreateStyleFont");
		pSnapWindowToEdge = FIND("SnapWindowToEdge");
		if ( pGetBBWnd == NULL || pMakeSticky == NULL
		     || pRemoveSticky == NULL || pGetSettingPtr == NULL
		     || pMakeStyleGradient == NULL || pCreateStyleFont ==NULL
		     || pSnapWindowToEdge == NULL )
		{
			return 0;
		}

		G_style = (*(tGetSettingPtr)pGetSettingPtr)(SN_TOOLBAR);

		SendMessage( (*((HWND(*)(void))pGetBBWnd))(),
		             BB_REGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)SUBSCRIBING_MSGS );

		make_always_on_top(hwnd, !0);
		(*(void (*)(HWND))pMakeSticky)(hwnd);
		} break;

	case WM_DESTROY:
		(*(void (*)(HWND))pRemoveSticky)(hwnd);
		make_always_on_top(hwnd, 0);
		SendMessage( (*((HWND(*)(void))pGetBBWnd))(),
		             BB_UNREGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)SUBSCRIBING_MSGS );
		break;

	case WM_NCHITTEST:
		return is_pressed(VK_CONTROL) ? HTCAPTION : HTCLIENT;

	case WM_WINDOWPOSCHANGING:
		_SnapWindowToEdge(pSnapWindowToEdge, (WINDOWPOS*)lp, 10, !0);
		break;

	case WM_PAINT: {
		HDC hdc;
		PAINTSTRUCT ps;
		RECT r;
		HDC buf;
		HBITMAP bmp;
		HGDIOBJ prev_bmp;
		HFONT style_font;
		HGDIOBJ prev_font;
		int bevel_width;
		int border_width;

		hdc = BeginPaint(hwnd, &ps);

		GetClientRect(hwnd, &r);
		buf = CreateCompatibleDC(NULL);
		bmp = CreateCompatibleBitmap(hdc, r.right, r.bottom);

		prev_bmp = SelectObject(buf, bmp);
		_MakeStyleGradient(pMakeStyleGradient, buf, &r, G_style, !0);

		style_font = _CreateStyleFont(pCreateStyleFont, G_style);
		prev_font = SelectObject(buf, style_font);
		SetBkMode(buf, TRANSPARENT);
		SetTextColor(buf, _StyleItem_TextColor(G_style));
		bevel_width = *(int*)(*(tGetSettingPtr)pGetSettingPtr)
		                                       (SN_BEVELWIDTH);
		border_width = *(int*)(*(tGetSettingPtr)pGetSettingPtr)
		                                        (SN_BORDERWIDTH);
		r.left += bevel_width + border_width;
		r.top += bevel_width + border_width;
		r.right -= bevel_width + border_width;
		r.bottom -= bevel_width + border_width;
		DrawText( buf, G_text, -1, &r,
		          DT_CENTER | DT_SINGLELINE | DT_VCENTER );
		SelectObject(buf, prev_font);

		BitBlt( hdc,
		        ps.rcPaint.left,
		        ps.rcPaint.top,
		        ps.rcPaint.right - ps.rcPaint.left,
		        ps.rcPaint.bottom - ps.rcPaint.top,
		        buf,
		        ps.rcPaint.left,
		        ps.rcPaint.top,
		        SRCCOPY );

		SelectObject(buf, prev_bmp);
		DeleteObject(style_font);
		DeleteObject(bmp);
		DeleteDC(buf);

		EndPaint(hwnd, &ps);
		} break;
	}

	return DefWindowProc(hwnd, msg, wp, lp);
}


static HWND
new_magnolia(HINSTANCE hinstance)
{
	WNDCLASS wc;
	HWND hwnd;

	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = magnolia_window_procedure;
	wc.hInstance = hinstance;
	wc.lpszClassName = MAGNOLIA_CLASSNAME;
	if (!RegisterClass(&wc)) return NULL;

	hwnd = CreateWindowEx( WS_EX_TOOLWINDOW,
	                       MAGNOLIA_CLASSNAME,
	                       MAGNOLIA_WINDOWNAME,
	                       WS_POPUP | WS_VISIBLE,
	                       0, 0, 24, 24,
	                       NULL,
	                       NULL,
	                       hinstance,
	                       NULL );
	if (hwnd == NULL) {UnregisterClass(MAGNOLIA_CLASSNAME, hinstance);}

	return hwnd;
}


static void
del_magnolia(HWND hwnd, HINSTANCE hinstance)
{
	DestroyWindow(hwnd);
	UnregisterClass(MAGNOLIA_CLASSNAME, hinstance);
}








/**
 * ==================
 *   Public section
 * ==================
 */


DLL_EXPORT LRESULT CALLBACK
mouse_procedure(int ncode, WPARAM wp, LPARAM lp)
{
	LRESULT result;
	int discard_message;
	const MOUSEHOOKSTRUCTEX* mhsex;
	event e;

	discard_message = 0;

	mhsex = (const MOUSEHOOKSTRUCTEX*)lp;

	/* must be here to avoid to miss changing state of buttons */
	check_mouse_button_state();

	if (ncode != HC_ACTION) goto END;

	e.e = event_id_from_windows_message((UINT)wp, mhsex);
	e.m = modifier_from_getkeystate();

	/* Check timeout */
		/* Just after a start event */
	if (G_state == STATE_GESTURING && gesture_stack_size() == 1) {
		DWORD current_time;

		current_time = GetTickCount();
		if (current_time < G_gesture_start_time) { /* 49.7days prob */
			G_gesture_start_time
				= G_gesture_start_time - current_time;
			current_time = MAXDWORD;
		}

		if ( G_option_timeout <= (int)( current_time
		                                - G_gesture_start_time) ) {
			event events[3];

			events[0] = *(gesture_stack_top());
				/* `e' can be an unplaybackable event,
				 * but it does not matter. */
			events[1] = e;
			events[2].e = EVENT_END;
			playback_event(events);

			change_state(STATE_IDLE, mhsex->hwnd);
			discard_message = !0;

			SendMessage(G_hwnd_magnolia, MAGNOLIA_TIMEOUT, 0, 0);
			lo(LO_DEBUG, "time out");
			goto END;
		}
	}

	/* Ignore the message which is not an event */
	if (e.e == EVENT_END) goto END;

	/* Avoid re-recognizing a playbacked gesture */
	if (0 < G_playbacking_gesture_num) {
		if ( G_playbacking_gesture[G_playbacking_gesture_pointer].e
		     == e.e )
		{
			G_playbacking_gesture_pointer++;

			if (!( G_playbacking_gesture_pointer
			       < G_playbacking_gesture_num ))
			{
				G_playbacking_gesture_num = 0;
			}

			goto END;
		}
	}

	/* Note that on_state_xxx may call on_state_yyy */
	if (G_state == STATE_IDLE)
		discard_message = on_state_idle(&e, mhsex->hwnd);
	else if (G_state == STATE_GESTURING)
		discard_message = on_state_gesturing(&e, mhsex->hwnd);
	else if (G_state == STATE_ACTION)
		discard_message = on_state_action(&e, mhsex->hwnd);
	else if (G_state == STATE_EPILOGUE)
		discard_message = on_state_epilogue(&e, mhsex->hwnd);


END:
	lo( LO_DEBUG | LO_DECONS,
	    "%2d %c %c %2d/%2d %c %c  %-18s  %04X/%-18s  X %4d/%4d  Y %4d/%4d"
	    "  hwnd %p  nc 0x%08X  ms%04X  md0x%08X",
	    gesture_stack_size(),
	    (discard_message ? 'D' : '_'),
	    (is_mouse_button_reversed() ? 'R' : 'N'),
	    G_playbacking_gesture_pointer,
	    G_playbacking_gesture_num,
	    "IGAE"[G_state],
	    (ncode == HC_ACTION ? 'A' : (ncode == HC_NOREMOVE ? 'N' : '_')),
	    message_id_to_string(wp),
	    e.m,
	    event_id_as_string(e.e),
	    mhsex->pt.x, G_cursor_point.x, mhsex->pt.y, G_cursor_point.y,
	    mhsex->hwnd, mhsex->wHitTestCode,
	    G_mouse_state,
	    mhsex->mouseData );

	result = CallNextHookEx(G_hhook, ncode, wp, lp);
	return discard_message ? !0 : result;
}




DLL_EXPORT int
beginPlugin(HINSTANCE hplugininstance)
{
	HOOKPROC hookproc;
	const char* func;

	lo(LO_DEBUG, "beginPlugin");

	if ( BBCmd_IsCompatible(BBCmd_VERSION, BBCmd_ICH_KNOW_INTERNAL)
	     != BBCmd_IC_COMPATIBLE )
	{
		lo(LO_ERR | LO_CONS, "Incompatible BBCmd");
		goto ERR_BBCMD;
	}

	G_pid_blackbox = GetCurrentProcessId();

	G_hinstance_self = hplugininstance;

	change_state(STATE_IDLE, NULL);
	if (!load_setting()) goto ERR_LOAD_SETTING;

	G_hwnd_flandre = new_flandre(hplugininstance);
	if (G_hwnd_flandre == NULL)
		{func = "new_flandre"; goto ERR_NF;}

	if (G_option_navigation_window) {
		G_hwnd_magnolia = new_magnolia(hplugininstance);
		if (G_hwnd_magnolia == NULL)
			{func = "new_magnolia"; goto ERR_NM;}
	}

	G_screen_width = GetSystemMetrics(SM_CXSCREEN);
	G_screen_height = GetSystemMetrics(SM_CYSCREEN);
	if (G_screen_width <= 0 || G_screen_height <= 0)
		{func = "GetSystemMetrics";  goto ERR_GETSM;}
	lo(LO_DEBUG, "w%d, h%d", G_screen_width, G_screen_height);

	hookproc = (HOOKPROC)GetProcAddress(hplugininstance,"mouse_procedure");
	if (hookproc == NULL) {func = "GetProcAddress"; goto ERR_GETPA;}

	G_hhook = SetWindowsHookEx(WH_MOUSE, hookproc, hplugininstance, 0);
	if (G_hhook == NULL) {func = "SetWindowsHookEx"; goto ERR_SETWHE;}

	lo(LO_INFO, "G_hhook: %p", G_hhook);

	return 0;

ERR_SETWHE:
ERR_GETPA:
ERR_GETSM:
ERR_NM:
	del_flandre(G_hwnd_flandre, hplugininstance);
ERR_NF:
	{
		LPVOID m;
		m = get_error_message(GetLastError());
		lo(LO_ERR | LO_CONS, "%s failed: %s", func, m);
		LocalFree(m);
	}
ERR_LOAD_SETTING:
	unload_setting();
ERR_BBCMD:
	return !0;
}


DLL_EXPORT void
endPlugin(HINSTANCE hplugininstance)
{
	lo(LO_DEBUG, "endPlugin");

	if (G_hhook != NULL) {
		if (UnhookWindowsHookEx(G_hhook) == 0) {
			LPVOID m;
			m = get_error_message(GetLastError());
			lo( LO_ERR | LO_CONS,
			    "UnhookWindowsHookEx(%p)failed: %s",
			    G_hhook, m );
			LocalFree(m);
		}
	}

	if (G_option_navigation_window)
		del_magnolia(G_hwnd_magnolia, hplugininstance);
	del_flandre(G_hwnd_flandre, hplugininstance);

	unload_setting();
}


DLL_EXPORT LPCSTR
pluginInfo(int field)
{
	switch (field) {
		case PLUGIN_NAME: return BBMOUSEGESTURE_NAME;
		case PLUGIN_VERSION: return BBMOUSEGESTURE_VERSION;
		case PLUGIN_AUTHOR: return BBMOUSEGESTURE_AUTHOR;
		case PLUGIN_RELEASE: return BBMOUSEGESTURE_RELEASE;
		case PLUGIN_LINK: return BBMOUSEGESTURE_LINK;
		case PLUGIN_EMAIL: return BBMOUSEGESTURE_EMAIL;
		default: return BBMOUSEGESTURE_DEFAULT;
	}
}




BOOL WINAPI
DllMain(HINSTANCE hinstance, DWORD reason, LPVOID __reserved__)
{
	if (reason == DLL_PROCESS_ATTACH) {
		G_pid_self = GetCurrentProcessId();
	}

	return TRUE;
}




/* __END__ */
